;; Generated package description from vdiff.el  -*- no-byte-compile: t -*-
(define-package "vdiff" "0.2.4.0.20210425.215531" "A diff tool similar to  vimdiff" '((emacs "24.4") (hydra "0.13.0")) :authors '(("Justin Burkett" . "justin@burkett.cc")) :maintainer '("Justin Burkett" . "justin@burkett.cc") :keywords '("diff") :url "https://github.com/justbur/emacs-vdiff")
